# funzione per fittare "mymixedtrees"

mymixedtree <- function(formula, dataset, random, subset = NULL,
                        family = binomial, tree.control = rpart.control(),
                        cv=TRUE, cpmin = 0.01, maxdepthtree=10, verbose = FALSE){
  
  # Parse formula
  Predictors <- paste(attr(terms(formula),"term.labels"),collapse="+")
  TargetName <- formula[[2]]
  
  # Subset the data if necessary
  if(identical(subset, NULL)){ 
    train <- dataset
  } else {
    train <- subset(dataset, dati[,subset])
  }
  
  ################
  # 1) TRAIN MODEL
  ################
  # Fit GLM (only fixed effect)
  
  GLM <- glm(formula = formula, data = train, family = family)
  if(verbose) {
    print(GLM)}
  hat.p = GLM$fitted.values
  
  # Fit regression tree on predicted probabilities
  # and prune tree (eventually)
  if(cv){
    CART <- rpart(formula = formula(paste(c("hat.p", Predictors), collapse = "~")), 
                  data = train, method = "anova", control = rpart.control(cp=cpmin, maxdepth=maxdepthtree))
    cventry <- which.min(CART$cptable[, "xerror"])
    xerrorcv <- CART$cptable[cventry, "xerror"]
    sexerrorcv <- xerrorcv + CART$cptable[cventry, "xstd"] 
    cpcvse <- CART$cptable[which.max(CART$cptable[, "xerror"] <= sexerrorcv), "CP"]
    CART.pr <- prune(CART, cp=cpcvse)
    if(verbose){
      print(CART.pr)
      rpart.plot(CART.pr)
    }
  } else {
    CART.pr <- rpart(formula = formula(paste(c("hat.p", Predictors), collapse = "~")), 
                     data = train, method = "anova", control = tree.control)
    if(verbose){
      print(CART.pr)
      rpart.plot(CART.pr)
    }
  }
  
  # get indicator of leaves
  leaves <- as.numeric(rownames(CART.pr$frame)[CART.pr$where])
  train[,"nodeInd"] <- leaves
  n.leaves = length(unique(leaves))
  
  # Fit GLMM 
  fixed = paste(c(TargetName, "as.factor(nodeInd)"), collapse = "~")
  GLMM <- glmer(formula = formula(paste(c(fixed, random[[2]]), collapse = "+")),
                data=train, family = family, control=glmerControl(optimizer="bobyqa"))
  if(verbose) {
    summary(GLMM)}
  
  #get fixed effect from the model and put it into tree
  adjtarg <- unique(cbind(leaves, predict(GLMM, re = NA, type = "response")))
  CART.pr$frame[as.character(adjtarg[,"leaves"]),]$yval <- adjtarg[,2]
  
  # fitted values
  fitted.GLMM <- predict(GLMM, type = "response")
  
  result <- list(
    Tree=CART.pr, 
    EffectModel=GLMM, 
    RandomEffects=ranef(GLMM),
    BetweenMatrix = VarCorr(GLMM)
  )
  class(result) <- "mymixedtree"
  return(result)
}


### Function: plot.mymixedtree
# This function plots the tree part of a mymixedtree tree
# Inputs:
#	x - a RE-EM tree object
plot.mymixedtree <- function(x, ...){
  require(rpart.plot)
  rpart.plot(x$Tree)
}

### Function: ranef.REEMtree
# This function extracts the vector of estimated random effects
ranef.mymixedtree <- function(object,...){
  return(object$RandomEffects)
}

predict.mymixedtree <- function(object, newdata, type = c("response", "matrix","terms"), ...){
  # get underlying tree structure
  tree.aux = object$Tree
  # substitute the row index as predicted value in each leaf
  tree.aux$frame$yval = as.numeric(rownames(tree.aux$frame))
  # get the leaf indicator for each newdata
  newdata[,"nodeInd"] <- predict(tree.aux, newdata= newdata)
  #predict using GLMM
  predict(object$EffectModel, newdata = newdata, type = type,...)
}

